/*
 * Copyright 2007, Intel Corporation
 *
 * This file is part of PowerTOP
 *
 * This program file is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program in a file named COPYING; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301 USA
 *
 * Authors:
 *	Arjan van de Ven <arjan@linux.intel.com>
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <sys/types.h>
#include <dirent.h>
#include <ncurses.h>
#include <time.h>
#include <wchar.h>

#include "powertop.h"

static WINDOW *title_bar_window;
static WINDOW *cstate_window;
static WINDOW *wakeup_window;
static WINDOW *battery_power_window;
static WINDOW *timerstat_window;
static WINDOW *suggestion_window;
static WINDOW *status_bar_window;

#define print(win, y, x, fmt, args...) do { if (dump) printf(fmt, ## args); else mvwprintw(win, y, x, fmt, ## args); } while (0)
#define p_wattrset(win, c) do { if (!dump ) wattrset(win, c); } while (0)
#define p_wattron(win, a) do { if (!dump ) wattron(win, a); } while (0)
#define p_wattroff(win, a) do { if (!dump ) wattroff(win, a); } while (0)
#define p_wbkgd(win, c) do { if (!dump ) wbkgd(win, c); } while (0)
#define p_werase(win) do { if (!dump ) werase(win); } while (0)
#define p_wrefresh(win) do { if (!dump ) wrefresh(win); } while (0)

char status_bar_slots[10][40];

static void cleanup_curses(void) {
	endwin();
}

static void zap_windows(void)
{
	if (title_bar_window) {
		delwin(title_bar_window);
		title_bar_window = NULL;
	}
	if (cstate_window) {
		delwin(cstate_window);
		cstate_window = NULL;
	}
	if (wakeup_window) {
		delwin(wakeup_window);
		wakeup_window = NULL;
	}
	if (battery_power_window) {
		delwin(battery_power_window);
		battery_power_window = NULL;
	}
	if (timerstat_window) {
		delwin(timerstat_window);
		timerstat_window = NULL;
	}
	if (suggestion_window) {
		delwin(suggestion_window);
		suggestion_window = NULL;
	}
	if (status_bar_window) {
		delwin(status_bar_window);
		status_bar_window = NULL;
	}
}


int maxx, maxy;

int maxtimerstats = 50;
int maxwidth = 200;

void setup_windows(void)
{
	int i, yline, maxcstateline = -2;
	for (i = 0; i <= MAX_CSTATE_LINES; i++) {
		if (strlen(cstate_lines[i]))
			maxcstateline++;
	}

	yline = (maxcstateline >= maxpstate) ? maxcstateline: maxpstate;

	/* number of states is one more than the MAX state! */
	if (!maxcstate && !maxpstate)
		yline = 0;
	else
		yline++;

	getmaxyx(stdscr, maxy, maxx);

	zap_windows();

	title_bar_window = subwin(stdscr, 1, maxx, 0, 0);
	cstate_window = subwin(stdscr, (yline + 3), maxx, 2, 0);
	wakeup_window = subwin(stdscr, 1, maxx, (yline + 4), 0);
	battery_power_window = subwin(stdscr, 2, maxx, (yline + 5), 0);
	timerstat_window = subwin(stdscr, maxy-16, maxx, (yline + 7), 0);
	maxtimerstats = maxy-16  -2;
	maxwidth = maxx - 18;
	suggestion_window = subwin(stdscr, 3, maxx, maxy-4, 0);
	status_bar_window = subwin(stdscr, 1, maxx, maxy-1, 0);

	strcpy(status_bar_slots[0], _(" Q - Quit "));
	strcpy(status_bar_slots[1], _(" R - Refresh "));

	p_werase(stdscr);
	refresh();
}

void initialize_curses(void)
{
	initscr();
	start_color();
	keypad(stdscr, TRUE);	/* enable keyboard mapping */
	nonl();			/* tell curses not to do NL->CR/NL on output */
	cbreak();		/* take input chars one at a time, no wait for \n */
	noecho();		/* dont echo input */
	curs_set(0);		/* turn off cursor */
	use_default_colors();

	init_pair(PT_COLOR_DEFAULT, COLOR_WHITE, COLOR_BLACK);
	init_pair(PT_COLOR_HEADER_BAR, COLOR_BLACK, COLOR_WHITE);
	init_pair(PT_COLOR_ERROR, COLOR_BLACK, COLOR_RED);
	init_pair(PT_COLOR_RED, COLOR_WHITE, COLOR_RED);
	init_pair(PT_COLOR_YELLOW, COLOR_WHITE, COLOR_YELLOW);
	init_pair(PT_COLOR_GREEN, COLOR_WHITE, COLOR_GREEN);
	init_pair(PT_COLOR_BLUE, COLOR_WHITE, COLOR_BLUE);
	init_pair(PT_COLOR_BRIGHT, COLOR_WHITE, COLOR_BLACK);

	atexit(cleanup_curses);
}

void show_title_bar(void)
{
	int i;
	int x;
	p_wattrset(title_bar_window, COLOR_PAIR(PT_COLOR_HEADER_BAR));
	p_wbkgd(title_bar_window, COLOR_PAIR(PT_COLOR_HEADER_BAR));
	p_werase(title_bar_window);

	print(title_bar_window, 0, 0,  "     PowerTOP version %s      (C) 2007 Intel Corporation", VERSION);

	p_wrefresh(title_bar_window);

	p_werase(status_bar_window);

	x = 0;
	for (i=0; i<10; i++) {
		if (strlen(status_bar_slots[i])==0)
			continue;
		p_wattron(status_bar_window, A_REVERSE);
		print(status_bar_window, 0, x, "%s", status_bar_slots[i]);
		p_wattroff(status_bar_window, A_REVERSE);
		x+= strlen(status_bar_slots[i])+1;
	}
	p_wrefresh(status_bar_window);
}

void show_c_and_p_states(void)
{
	int i, count = 0;

	p_werase(cstate_window);

	for (i = 0; i <= MAX_CSTATE_LINES; i++) {
		if (i == topcstate+1)
			p_wattron(cstate_window, A_BOLD);
		else
			p_wattroff(cstate_window, A_BOLD);
		if (strlen(cstate_lines[i])) {
			print(cstate_window, count, 0, "%s", cstate_lines[i]);
			count++;
		}
	}

	for (i = 0; i <= (dump ? MAX_NUM_PSTATES : maxpstate + 1); i++) {
		if (i == topfreq+1)
			p_wattron(cstate_window, A_BOLD);
		else
			p_wattroff(cstate_window, A_BOLD);
		print(cstate_window, i, 38, "%s", cpufreqstrings[i]);
	}

	p_wrefresh(cstate_window);
}

void show_cstates(void)
{

	if (!maxcstate) {
		strcpy(cstate_lines[0], "<C-state information is unavailable>\n");
		strcpy(cstate_lines[1], "");
		topcstate = 99; /* Dummy value, to avoid "bold" text */
	}

	if (!maxpstate) {
		strcpy(cpufreqstrings[0],
		       "<P-state information is unavailable>\n");
		strcpy(cpufreqstrings[1], "");
		topfreq = 99; /* Dummy value, to avoid "bold" text */
	}

	show_c_and_p_states();
}

void show_acpi_power_line(double rate, double cap, double capdelta, time_t ti)
{
	char buffer[1024];

	sprintf(buffer,  _("no ACPI power usage estimate available") );

	p_werase(battery_power_window);
	if (rate > 0.001) {
		char *c;
		sprintf(buffer, _("Power usage (ACPI estimate): %3.1fW (%3.1f hours)"), rate, cap/rate);
		strcat(buffer, " ");
		c = &buffer[strlen(buffer)];
		if (ti>180 && capdelta > 0)
			sprintf(c, _("(long term: %3.1fW,/%3.1fh)"), 3600*capdelta / ti, cap / (3600*capdelta/ti+0.01));
	}
	else if (ti>120 && capdelta > 0.001)
		sprintf(buffer, _("Power usage (5 minute ACPI estimate) : %5.1f W (%3.1f hours left)"), 3600*capdelta / ti, cap / (3600*capdelta/ti+0.01));

	print(battery_power_window, 0, 0, "%s\n", buffer);
	p_wrefresh(battery_power_window);
}

void show_pmu_power_line(unsigned sum_voltage_mV,
                         unsigned sum_charge_mAh, unsigned sum_max_charge_mAh,
                         int sum_discharge_mA)
{
	char buffer[1024];

	if (sum_discharge_mA != 0)
	{
		unsigned remaining_charge_mAh;

		if (sum_discharge_mA < 0)
		{
			/* we are currently discharging */
			sum_discharge_mA = -sum_discharge_mA;
			remaining_charge_mAh = sum_charge_mAh;
		}
		else
		{
			/* we are currently charging */
			remaining_charge_mAh = (sum_max_charge_mAh
						- sum_charge_mAh);
		}

		snprintf(buffer, sizeof(buffer),
			 _("Power usage: %3.1fW (%3.1f hours)"),
			 sum_voltage_mV * sum_discharge_mA / 1e6,
			 (double)remaining_charge_mAh / sum_discharge_mA);
	}
	else
		snprintf(buffer, sizeof(buffer),
			 _("no power usage estimate available") );

	p_werase(battery_power_window);
	print(battery_power_window, 0, 0, "%s\n", buffer);
	p_wrefresh(battery_power_window);
}


void show_wakeups(double d, double interval, double C0time)
{
	p_werase(wakeup_window);

	p_wbkgd(wakeup_window, COLOR_PAIR(PT_COLOR_RED));
	if (d <= 25.0)
		p_wbkgd(wakeup_window, COLOR_PAIR(PT_COLOR_YELLOW));
	if (d <= 10.0)
		p_wbkgd(wakeup_window, COLOR_PAIR(PT_COLOR_GREEN));

	/*
	 * if the cpu is really busy.... then make it blue to indicate
	 * that it's not the primary power consumer anymore
	 */
	if (C0time > 25.0)
		p_wbkgd(wakeup_window, COLOR_PAIR(PT_COLOR_BLUE));

	p_wattron(wakeup_window, A_BOLD);
	print(wakeup_window, 0, 0, _("Wakeups-from-idle per second : %4.1f\tinterval: %0.1fs\n"), d, interval);
	p_wrefresh(wakeup_window);
}

void show_timerstats(int nostats, int ticktime)
{
	int i;
	p_werase(timerstat_window);

	if (!nostats) {
		int counter = 0;
		print(timerstat_window, 0, 0, _("Top causes for wakeups:\n"));
		for (i = 0; i < linehead; i++)
			if ((lines[i].count > 0 || lines[i].disk_count > 0) && counter++ < maxtimerstats) {
				char c = ' ';
				if (lines[i].disk_count)
					c = 'D';
				if ((lines[i].count * 1.0 / ticktime) >= 10.0)
					p_wattron(timerstat_window, A_BOLD);
				else
					p_wattroff(timerstat_window, A_BOLD);
				if (showpids)
					print(timerstat_window, i+1, 0," %5.1f%% (%5.1f)%c  [%6s] %s\n", lines[i].count * 100.0 / linectotal,
						lines[i].count * 1.0 / ticktime, c,
						lines[i].pid, lines[i].string);
				else
					print(timerstat_window, i+1, 0," %5.1f%% (%5.1f)%c  %s\n", lines[i].count * 100.0 / linectotal,
						lines[i].count * 1.0 / ticktime, c,
						lines[i].string);
				}
	} else {
		if (geteuid() == 0) {
			print(timerstat_window, 0, 0, _("No detailed statistics available; please enable the CONFIG_TIMER_STATS kernel option\n"));
			print(timerstat_window, 1, 0, _("This option is located in the Kernel Debugging section of menuconfig\n"));
			print(timerstat_window, 2, 0, _("(which is CONFIG_DEBUG_KERNEL=y in the config file)\n"));
			print(timerstat_window, 3, 0, _("Note: this is only available in 2.6.21 and later kernels\n"));
		} else
			print(timerstat_window, 0, 0, _("No detailed statistics available; PowerTOP needs root privileges for that\n"));
	}


	p_wrefresh(timerstat_window);
}

void show_suggestion(char *sug)
{
	p_werase(suggestion_window);
	print(suggestion_window, 0, 0, "%s", sug);
	p_wrefresh(suggestion_window);
}
